/* T2.C */

/*********************************************************************************
License for SLC - Tibetan Calendar software for Sherab Ling

Copyright (c) 2009, 2010 Edward Henning

Permission is hereby granted, free of charge, to any person  obtaining a copy of
this software and associated documentation files (the "Software"), to deal in the
Software without restriction, including without limitation the rights to use, 
copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
Software, and to permit persons to whom the Software is furnished to do so, subject 
to the following conditions: 

The above copyright notice and this permission notice shall be included in all copies
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
************************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <conio.h>
#include "tc.h"
#include "tc.ext"
#include "bcd.h"
#include "bcd.ext"

// Local to this file only:

long int  spz_c, spz_b, spz_f, spz_j;

long int gzada[6];
long int nyida[6];
long int nyidm[6], gzadm[6], tsedm[6], nyilm[6];

static long int nyidmg[6] = { 2, 10, 58, 1, 17, 0 };
static long int gzadmg[6] = { 1, 31, 50, 0, 480, 0 };
static long int tsedmg[6] = { 0, 59, 3, 4, 16, 0 };
static long int nyilmg[6] = { 0, 4, 21, 5, 43, 0 }; // Each lunar day

static long int nyidmb[6] = { 2, 10, 58, 2, 10, 0 };
static long int gzadmb[6] = { 1, 31, 50, 0, 0, 0 };
static long int tsedmb[6] = { 0, 59, 3, 4, 0, 0 };
static long int nyilmb[6] = { 0, 4, 21, 5, 9, 0 };

static long int nyifac[6] = { 6, 45, 0, 0, 0, 0 };
static long int nyihaf[6] = { 13, 30, 0, 0, 0, 0 };
static long int nyibye[6] = { 4, 1, 1, 4, 6, 6};
static long int nyidom[6] = { 6, 10, 11, 10, 6, 0};

static long int gzabye[14] = { 5, 5, 4, 3, 2, 1, -1, -2, -3,
                                                  -4, -5, -5, -5, 5};
static long int gzadom[14] = { 5, 10, 15, 19, 22, 24, 25, 24, 22,
                                                   19, 15, 10, 5, 0};

/**** Figures for Bir Tantra "grub rtsis", year 1987, new (1) ****/

static long int gda1[6] = { 1, 42, 47, 3, 465, 0 }; //
static long int nda1[6] = { 25, 41, 58, 2, 176930, 0 }; //

/*** ROUTINE TO SET FOR GRUB RTSIS ****/

void set_grub ( void )
  {
    int  i;
    for ( i = 0; i < 6; ++i )
      nyidm[i] = nyidmg[i];
    for ( i = 0; i < 6; ++i )
      gzadm[i] = gzadmg[i];
    for ( i = 0; i < 6; ++i )
      tsedm[i] = tsedmg[i];
    for ( i = 0; i < 6; ++i )
      nyilm[i] = nyilmg[i];
    sun_f = 67L;
    gza_f = 707L;
  } // END - set_grub ()

/*** ROUTINE TO SET FOR BYED RTSIS ****/

void set_byed ( void )
  {
    int  i;
    for ( i = 0; i < 6; ++i )
      nyidm[i] = nyidmb[i];
    for ( i = 0; i < 6; ++i )
      gzadm[i] = gzadmb[i];
    for ( i = 0; i < 6; ++i )
      tsedm[i] = tsedmb[i];
    for ( i = 0; i < 6; ++i )
      nyilm[i] = nyilmb[i];
    sun_f = 13L;
    gza_f = 13L;   
  } // END - set_byed ()

// Routine to set epoch parameters 

void set_epoch ( void )
  {
    char chr;
    int  i;
    cls ();
    pos ( 0, 0 );
    printf ( "Select from the following:\n\n" );
    for ( i = 0; i < 5; ++i )
      printf ( "%s\n", e_str[i] );
//    do
//      chr = getch ();
//    while ( chr < '1' || chr > '5' );
//    if ( chr >= '1' && chr <= '9' )
//      epch = chr - '0' - 1;

// Only the 1987 epoch currently implemented.
    
    chr = '1';
    epch = 0;
    
    if ( chr == '1' )   // Sherab Ling Tantra, new - 1987
      {
        set_grub ();
        for ( i = 0; i < 6; ++i )
          gzada[i] = gda1[i];
        for ( i = 0; i < 6; ++i )
          nyida[i] = nda1[i];
        epch_yr = 1987L;
        eyr_a = 38L;   // Lower part of month calculation. // OK
        ril_a = 19L; // OK
        ril_b = 48L; // OK
        spz_b = 18L; // 64 fraction - OK
        spz_c = 250L; // 707 fraction - OK
        spz_f = 1L;   // OK
        spz_j = 2446884L;  //
        nyi_cnt = 86548L;

        rahupart = 8; // 230 fraction for Rahu cycle //

        meradd = 6104L; // Figures to be added for planets
        venadd = 1561L; // nyin zhag dal ba. These are days of cycle. //
        maradd = 94L; //
        jupadd = 4105L; //
        satadd = 6867L; //
        dragkadd = 0L; // ????  This is subtracted for drag po'i rkang 'dzin. NOT USED

// OVER-RIDE PREVIOUS SETTINGS:
        sun_f = 4815377L;
        nyidm[3] = 2L;   // These correct solar motion.
        nyidm[4] = 3846950L;
//      tsedm[2] = 4L; // TEST, maybe they do this???
//      tsedm[3] = 0L; // TEST, maybe they do this???
//      tsedm[4] = 0L; // TEST, maybe they do this???
// Motion each lunar day:
        for ( i = 0; i < 6; ++i )
          nyilm[i] = nyidm[i];
        div_g6 ( nyilm, 30L, sun_f, 1L );

        zlasho1 = 16L; // Maybe 17 & 18 are better??
        zlasho2 = 17L;

        zlapure = 0;
        tsurlug = 0;
      }
    } // END - set_epoch ()

// ROUTINE TO CALCULATE POSITIONS OF RAHU

void do_rahu ( long int m, long int tt ) // KTC 96
  {
    long int t, mth;

    mth = ( m + rahupart ) % 230L;
    t = ( ( m + rahupart ) % 230L ) * 30L + 15L;
    rahutsa[0] = 0L;
    rahutsa[1] = 0L;
    rahutsa[2] = 14L;
    rahutsa[3] = 0L;
    rahutsa[4] = 12L;

    mul_gen ( rahutsa, rahutsa, t, 27L, 23L );
    clear_a_b ();
    lista[0] = 27L;
    sub_gen ( rahudong, lista, rahutsa, 27L, 23L );
    listb[0] = 13L;
    listb[1] = 30L;
    add_gen ( rahujug, rahudong, listb, 27L, 23L );

    rahutsa[0] = 0L;
    rahutsa[1] = 0L;
    rahutsa[2] = 14L;
    rahutsa[3] = 0L;
    rahutsa[4] = 12L;
    mul_gen ( rahutsa, rahutsa, t + 15L, 27L, 23L ); // For new Moon.
    clear_a_b ();
    lista[0] = 27L;
    sub_gen ( rahudong30, lista, rahutsa, 27L, 23L );
    listb[0] = 13L;
    listb[1] = 30L;
    add_gen ( rahujug, rahudong, listb, 27L, 23L );

    sprintf ( printstring,
    "Rahu month: %ld, gdong, 15th: %ld;%ld,%ld,%ld,%ld - 30th: %ld;%ld,%ld,%ld,%ld", 
mth, rahudong[0], rahudong[1], rahudong[2], rahudong[3], rahudong[4], 
rahudong30[0], rahudong30[1], rahudong30[2], rahudong30[3], rahudong30[4] );
    printf ( "%s\n", printstring );
    if ( printinprog )
      prin_str ( printstring );
  } // END - do_rahu ()

// Function to calculate true month, "zla ba rnam par dag pa"
void zla_dag ( long int y, long int m ) // KTC 16
  {
    long int   yr, a, b; // c;
    yr = y - epch_yr;
    
// Need to use this: m x 1;287179 + 0;5566410 (1,9343575)    
    
    a = 12L * yr + m - 3L;
    if ( a >= 0L )
      {
        b = 2L * a + eyr_a;
        zladag[1] = b % 65L;
        zladag[0] = a + b / 65L;
      }
    else
      {
        zladag[1] = 0L;
        zladag[0] = 0L;
      }
  } // END - zla_dag ()

// Adjust moth number, if needed
void adj_zla ( void )
  {
    if ( zladag[1] == zlasho1 || zladag[1] == zlasho2 )
      adj_mth = -tm;
    else if ( zladag[1] > zlasho2 )
      adj_mth = tm - 1L;
    else if ( zladag[1] == 0L || zladag[1] == 1L )
      {
        if ( !zeromthfg )
          {
            zeromthfg = 1;
            adj_mth = tm - 1L;
            zladag[0] = zladag[0] - 1L;
          }
        else
          {
            zeromthfg = 0;
            adj_mth = tm;
            zladag[0] = zladag[0] + 1L;
          }
      }
    else  // Arrive here if 1 < zladag[1] < zlasho1
      adj_mth = tm;
    if ( adj_mth == 0L )
      adj_mth = 12L;
  } // - adj_zla ()

// Function to calculate lunar anomaly, "ril cha"
void rilchaf ( long int x ) // KTC 21
  {
    long int   a, b;
    b = x + ril_b;
    a = 2L * x + ril_a + b / 126L;
    rilcha[1] = b % 126L;
    rilcha[0] = a % 28L;
  } // END - void rilchaf ( long int x )

// Function to calculate general day, "spyi zhag"
// Calculates Julian for the integer Julian day count at noon UT
// on weekday of "gza' dag". Put into juldat.
void spi_zagf ( void ) // KTC 46
  {
    long int b, c;
    spizag = cur_mth * 30L + tt;
    c = spizag + spz_c;
    spz_frac_c = c % 707L; 
    b = spizag + spz_b;
    b = b + c / 707L;
    spz_frac_b = b % 64L;  
    spizag = spizag - b / 64L;
    c = ( spizag + spz_f ) % 7L;
    b = gzadag[0];
    if ( c != b )
      {
        if ( c > 4L && b < 2L )
          b += 7L;
        else if ( b > 4L && c < 2L )
          c += 7L;
        spizag = spizag + b - c;
      }
    if ( b - c > 2L )
      printf ( "\nERROR IN GENERAL DAY ROUTINE: %ld\n", b - c );
    juldat = spizag + spz_j;
    jul2date ( juldat );
  } // - spi_zagf ();

long int gregd2jul ( long int D, long int M, long int Y )
{
  long int JD;
  int calndar;  // Gregorian = 1.

// In change from Julian to Gregorian calendars, in 1582, Oct 4th was followed
// by Oct 15th

  if ( Y > 1582L )
    calndar = 1;
  else if ( Y < 1582L )
    calndar = 0;
  else
    {
      if ( M > 10L )
        calndar = 1;
      else if ( M < 10L )
        calndar = 0;
      else
        {
          if ( D >= 15L )
            calndar = 1;
          else
            calndar = 0;
        }
    }
  if ( calndar )  // Gregorian:
    {

// Seidelmann's formula (2). This routine is valid between March 1, 1600
// through Feb 28, 2100. Calculates Julian date starting at noon on the date in
// question.

//JD = D + ( 1461 * ( Y + ( ( M - 14L ) / 12L ) ) ) / 4L
//       + ( 367L * ( M - 2L - 12L * ( ( M - 14L ) / 12L ) ) ) / 12L
//       + ( 24002L - ( 12L * Y ) - M ) / 1200L + 1721073L + 1L;
// The 1 is added because these routines calculate for noon.

// CHANGED - the following is from ESAA, 1992, p. 604.

  JD = D + ( 1461 * ( Y + 4800L + ( M - 14L ) / 12L ) ) / 4L
         + ( 367L * ( M - 2L - 12L * ( ( M - 14L ) / 12L ) ) ) / 12L
         - ( 3L * ( ( Y + 4900L + ( M - 14 ) / 12L ) / 100L ) ) / 4L - 32075L;

    }
  else  // Julian:
    {
      JD = 367L * Y - ( 7L * ( Y + 5001L + ( M - 9L ) / 7L )) / 4L
           + ( 275L * M ) / 9L + D + 1729777L;
    }

  return ( JD );
} // END - gregd2jul ()

void jul2date ( long int jd )
  {
    long int l, n, j, k, i;

// This algorithm is from photcopied notes, from James Neely.
// Also, checked with ESAA, 1992, p. 604
// Calculates date, at noon on which the Julian date starts.
// Julian 0 starts Greenwich mean noon on 1st Jan 4713 BC, Julian proleptic
// calendar.
// In change from Julian to Gregorian calendars, in 1582, Oct 4th was followed
// by Oct 15th

// First, get day of week:

  doweek = jd - 7L * (( jd + 1 ) / 7L ) + 2L;
  if ( doweek == 7L )
    doweek = 0L;
  if ( doweek > 7L )
    {
      printf ( "ERROR IN DAY OF WEEK ROUTINE:\n" );
      getch ();
    }

  if ( jd >= 2299161L )  // Gregorian calendar:
    { // This has been tested between March 1, 1600 and Jan 31, 2100
    l = jd + 68569L;
    n = ( 4L * l ) / 146097L;
    l = l - ( 146097L * n + 3L ) / 4L;

//  wy = 4000L * ( l + 1L ) / 1461001L;

    l2bcd ( bcda, 4000L );
    mulbcdl ( bcda, bcda, l + 1L );
    divbcdl ( bcda, bcda, 1461001L );
    wy = bcd2l ( bcda );

    l = l - ( 1461L * wy ) / 4L + 31L;
    wm = ( 80L * l ) / 2447L;
    wd = l - ( 2447L * wm ) / 80L;
    l = wm / 11L;
    wm = wm + 2L - 12L * l;
    wy = 100L * ( n - 49L ) + wy + l;
// j = month, k = day, i = year
    }
  else // Julian calendar
    {
      j = jd + 1402L;
      k = ( j - 1L ) / 1461L;
      l = j - 1461L * k;
      n = ( l - 1L ) / 365L - l / 1461L;
      i = l - 365L * n + 30L;
      j = ( 80L * i ) / 2447L;
      wd = i - ( 2447L * j ) / 80L;
      i = j / 11L;
      wm = j + 2L - 12L * i;
      wy = 4L * k + n + i - 4716L;
    }
  } // END - jul2date ()

// Function to calculate month mean Sun, "nyi ma'i dhru ba".
void nyi_dru ( long int x )
  {
    mul_gen ( nyidru, nyidm, x, 27L, sun_f );
    add_gen ( nyidru, nyidru, nyida, 27L, sun_f );
  } // END - nyi_dru ()

// Function to calculate month mean weekday, "gza'i dhru ba".

void gza_dru ( long int x )
  {
    mul_gen ( gzadru, gzadm, x, 7L, gza_f );
    add_gen ( gzadru, gzadru, gzada, 7L, gza_f );
  } // END - gza_dru ()

// Function to calculate lunar day solar longitude, "nyi ma'i longs spyod".
void nyi_lon ( long int x )
  {
    mul_gen ( nyilon, nyilm, x, 27L, sun_f );
  } // END - nyi_lon ()

// Function to calculate lunar day mean weekday, "tshes kyi dhru ba". 
void tse_dru ( long int x )
  {
    mul_gen ( tsedru, tsedm, x, 7L, gza_f );
  } // END - tse_dru ()

// Function to calculate true solar longitude, "nyi dag". 
void nyi_dag ( long int *a1 )
  {
    long int test, tquo, trem;

    clear_a_b ();
    listb[0] = 1L;
    listb[1] = 43L;
    listb[2] = 30L;
    add_gen ( lista, a1, listb, 27L, sun_f );
    sub_gen ( nyiwor, lista, nyifac, 27L, sun_f );
    test = 60L * nyiwor[0] + nyiwor[1];
    if ( test < 810L )
      nyidor = 0;
    else
      {
        nyidor = 1;
        sub_gen ( nyiwor, nyiwor, nyihaf, 27L, sun_f);
        test = 60L * nyiwor[0] + nyiwor[1];
      }
    trem = test % 135;
    tquo = test / 135;
    if ( tquo == 0 )
      tquo = 6;

    clear_a_b ();

    lista[2] = ( trem * 60L + nyiwor[2] ) * nyibye[(int)tquo - 1];
    lista[3] = nyiwor[3] * nyibye[(int)tquo - 1];
    lista[4] = nyiwor[4] * nyibye[(int)tquo - 1];

    div_g6 ( lista, 135L, sun_f, 1L );

    clrlst (zerlst);
    add_gen ( lista, zerlst, lista, 27L, sun_f );

    listb[1] = nyidom[(int)tquo - 1];
    if ( tquo == 3L || tquo == 4L || tquo == 5L ) // Then, subtract:
      sub_gen ( listc, listb, lista, 27L, sun_f );
    else
      add_gen ( listc, listb, lista, 27L, sun_f );

    if ( nyidor == 0 )
      sub_gen ( nyidag, a1, listc, 27L, sun_f );
    else
      add_gen ( nyidag, a1, listc, 27L, sun_f );

// What about this??? Old routine from TC.EXE - of any use now???

    if ( epch == 12 ) // Sherab Ling - ONLY for present calendar
      { // Now set up listc to be used for the correction to the weekday:
        clrlst (zerlst);

        sub_gen ( nyiwor, a1, nyifac, 27L, sun_f );

        test = 60L * nyiwor[0] + nyiwor[1];
        if ( test < 810L )
          nyidor = 0;
        else
          {
            nyidor = 1;
            sub_gen ( nyiwor, nyiwor, nyihaf, 27L, sun_f);
//          printf ( "Nyiwor: " );
//          prn_lst ( nyiwor );
            test = 60L * nyiwor[0] + nyiwor[1];
          }
        trem = test % 135;
        tquo = test / 135;
        if ( tquo == 0 )
          tquo = 6;

//  printf ( "Trem = %ld, Tquo = %ld\n", trem, tquo );

        clear_a_b ();

        lista[2] = ( trem * 60L + nyiwor[2] ) * nyibye[(int)tquo - 1];
        lista[3] = nyiwor[3] * nyibye[(int)tquo - 1];
        lista[4] = nyiwor[4] * nyibye[(int)tquo - 1];

//  printf ( "Coefficient = %ld\n", nyibye[(int)tquo - 1] );

        div_g6 ( lista, 135L, sun_f, 1L );

        clrlst (zerlst);
        add_gen ( lista, zerlst, lista, 27L, sun_f );

        listb[1] = nyidom[(int)tquo - 1];

//  printf ( "Step value = %ld\n", listb[1] );

        if ( tquo == 3L || tquo == 4L || tquo == 5L ) // Then, subtract:
          {
            sub_gen ( listc, listb, lista, 27L, sun_f );
          }
        else
          {
            add_gen ( listc, listb, lista, 27L, sun_f );
          }
      } // END of --- if ( epch == 12) // Sherab Ling      
          
  } // END - nyi_dag ()

// Function to calculate true weekday, "gza' dag".
void gza_dag ( long int *a1 )
  {
    long int tot, trem, chasha, rilpo;
    long int gzawor[5];

    clrlst (list1);
    clrlst (list2);
    clrlst (zerlst);
    rilpo = rilcha[0] + tt;
    trem = rilpo % 14;
    if ( trem == 0 )
      trem = 14;
    list2[1] = gzadom[(int)trem - 1];
//  gza_short_flg = 1; 
    if ( gza_short_flg ) // Flag for less accurate weekday adjustment
      {
        chasha = rilcha[1] * gzabye[(int)trem - 1L] * 254520L; //60*6*707
        // Maximum value = 125 * 5 * 254520 = 159075000 - OK
        tot = chasha / 126L;
      }
    else
      {
        l2bcd ( bcda, ( 30L * rilcha[1] + tt ) * gzabye[(int)trem - 1] );
        mulbcdl ( bcda, bcda, 360L * gza_f );
        divbcdl ( bcda, bcda, 3780L );
        tot = bcd2l (bcda);
      }
    if ( tot < 0 )
      {
        list1[4] = -tot;
        add_gen ( list1, list1, zerlst, 7L, gza_f );
        sub_gen ( list3, list2, list1, 7L, gza_f );
      }
    else
      {
        list1[4] = tot;
        add_gen ( list1, list1, zerlst, 7L, gza_f );
        add_gen ( list3, list2, list1, 7L, gza_f );
      }
    if (( rilpo / 14 ) % 2 == 0 )
      add_gen ( gzawor, a1, list3, 7L, gza_f);
    else
      sub_gen ( gzawor, a1, list3, 7L, gza_f);

    l2bcd ( bcda, gzawor[4] );
    mulbcdl ( bcda, bcda, sun_f );
    divbcdl ( bcda, bcda, gza_f );
    gzawor[4] = bcd2l ( bcda );

//  The above replaces:
//  gzawor[4] = ( sun_f * gzawor[4] ) / gza_f;

    if ( nyidor == 0 )
      sub_gen ( gzadag, gzawor, listc, 7L, sun_f );
    else
      add_gen ( gzadag, gzawor, listc, 7L, sun_f );

// Convert back the lowest fractional part:

    l2bcd ( bcda, gzadag[4] );
    mulbcdl ( bcda, bcda, gza_f );
    divbcdl ( bcda, bcda, sun_f );
    gzadag[4] = bcd2l ( bcda );
  } // END - gza_dag ()
